#include "pixelpipeline.h"
#include <framebuffer.h>
#include <texture.h>
#include <buffer.h>
#include <shader.h>
#include <cstring>
#include <glm/vec2.hpp>
#include <qtutils.h>

#define SCROLL_SPEED 0.998f

// MESH (2D points)

class PixelMesh
{
private:
   GLuint m_vao;
   
   void updateBuffer()
   {
      TBuffer<Pix>::BufferEditor editor(pixBuffer);
      std::memcpy(editor.getPointer(), m_pixels.data(), m_pixels.size()*sizeof(Pix));
      m_pixels.clear();
   }
   
public:
   struct Pix
   {
      glm::vec2 pos;
      glm::vec3 color;
   };
   
   std::vector<Pix> m_pixels;
   TBuffer<Pix> *pixBuffer;
   
   PixelMesh()
   {
      m_pixels.push_back({glm::vec2(0), glm::vec3(0)}); // buffer must not be empty
      glGenVertexArrays(1, &m_vao);
      glBindVertexArray(m_vao);
      
      pixBuffer = new TBuffer<Pix>(m_pixels, Buffer::VBO, true);
      pixBuffer->setVertexAttrib(0, 2);
      pixBuffer->setVertexAttrib(1, 3, sizeof(glm::vec2));
      
      glBindVertexArray(0);
   }
   
    ~PixelMesh()
    {
        if(m_vao != 0)
        {
            delete(pixBuffer);
            glDeleteVertexArrays(1, &m_vao);
            m_vao = 0;
        }
    }
   
   void draw()
   {
      glBindVertexArray(m_vao);
      updateBuffer();
      glDrawArrays(GL_POINTS, 0, m_pixels.size());
      glBindVertexArray(0);
   }
};

// PIPELINE

PixelPipeline::PixelPipeline(int mapWidth, int mapHeight) :
   m_mapWidth(mapWidth),
   m_mapHeight(mapHeight),
   m_camera(0, 0, 1)
{
    m_targetFBO = FrameBuffer::screen;
    m_mapFBO = new FrameBuffer();
    m_mapTex = new Texture(GL_RGBA, GL_RGBA, m_mapWidth, m_mapHeight, GL_UNSIGNED_BYTE, GL_TEXTURE_RECTANGLE);
    m_mapTex->setFiltering(GL_NEAREST);
    m_mapTex->setWrap(GL_CLAMP_TO_EDGE);
    m_mapFBO->addTexture(m_mapTex, GL_COLOR_ATTACHMENT0);
    m_mapFBO->initColorAttachments();

    // set up vao
    const float vertices[] = {
       -1.0f, -1.0f,
        3.0f, -1.0f,
       -1.0f,  3.0f
    };
    glGenVertexArrays(1, &m_vao);
    glBindVertexArray(m_vao);
    glGenBuffers(1, &m_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, m_vbo);
    glBufferData(GL_ARRAY_BUFFER, 3 * 2 * sizeof(GLfloat), vertices, GL_STATIC_DRAW);
    glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, sizeof(GLfloat)*2, NULL);
    glEnableVertexAttribArray(0);

    std::string vertSource = QtUtils::fileToString(":shaders/shaders/pixel.vert.glsl").toStdString();
    std::string fragSource = QtUtils::fileToString(":shaders/shaders/pixel.frag.glsl").toStdString();
    m_texMapShader = new Shader(vertSource, fragSource);
    vertSource = QtUtils::fileToString(":shaders/shaders/world.vert.glsl").toStdString();
    fragSource = QtUtils::fileToString(":shaders/shaders/world.frag.glsl").toStdString();
    m_renderShader = new Shader(vertSource, fragSource);
    m_changes = new PixelMesh();
}

void PixelPipeline::pushChange(const glm::vec2 &pos, const glm::vec3 &color)
{
    m_changes->m_pixels.push_back({pos, color});
}

void PixelPipeline::updateChanges()
{
    if(!m_changes->m_pixels.empty())
    {
        m_mapFBO->bindFBO();
        glViewport(0, 0, m_mapWidth, m_mapHeight);
        m_texMapShader->bind();
        m_changes->draw();
    }
}

void PixelPipeline::renderGL(Scene *scene)
{
    m_targetFBO->bindFBO();
    glViewport(0, 0, m_width, m_height);
    glDisable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);
    m_renderShader->bind();
    m_renderShader->bindVec3(m_renderShader->getLocation("camera"), m_camera);
    m_renderShader->bindVec2(m_renderShader->getLocation("worldSize"), glm::vec2(m_mapWidth, m_mapHeight));
    m_renderShader->bindVec2(m_renderShader->getLocation("screenSize"), glm::vec2(m_width, m_height));
    m_mapTex->bind(0);
    m_renderShader->bindInteger(m_renderShader->getLocation("colorMap"), 0);

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glBindVertexArray(m_vao);
    glDrawArrays(GL_TRIANGLES, 0, 3);
    glBindVertexArray(0);
}

void PixelPipeline::resizeGL(int w, int h)
{
    m_width = w;
    m_height = h;
}

void PixelPipeline::cameraZoom(int nbScrolls)
{
    while(nbScrolls != 0)
        {
            if(nbScrolls > 0)
            {
                m_camera.z *= SCROLL_SPEED;
                --nbScrolls;
            }
            else
            {
                m_camera.z /= SCROLL_SPEED;
                ++nbScrolls;
            }
        }
}
